import argparse
import os
from pathlib import Path
from .orchestration import Orchestrator


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('--run_name',
                        type=str,
                        help='name of the folder which contains config file',
                        required=True)

    parser.add_argument('--data_dir',
                        type=str,
                        help='absolute path to directory of the dataset',
                        required=False)
    parser.add_argument('--project_root',
                        type=str,
                        help='absolute path to the root of your project(repository)',
                        required=False)

    return parser.parse_args()


def main():
    args = parse_args()

    data_dir = args.data_dir
    if data_dir is not None:
        data_dir = Path(data_dir)
    run_name = str(args.run_name)
    # project_root = Path(args.project_root)
    project_root = args.project_root
    if project_root is None:
        project_root = Path(__file__).absolute().parent.parent
    else:
        project_root = Path(project_root)

    mlflow_tracking_uri = os.getenv('MLFLOW_TRACKING_URI')
    if mlflow_tracking_uri is None:
        raise Exception('set MLFLOW_TRACKING_URI as an environmental variable.')
    else:
        mlflow_tracking_uri = str(mlflow_tracking_uri)

    eval_reports_dir = os.getenv('EVAL_REPORTS_DIR')
    if eval_reports_dir is None:
        raise Exception('set EVAL_REPORTS_DIR as an environmental variable.')
    else:
        eval_reports_dir = Path(eval_reports_dir)

    orchestrator = Orchestrator(run_name=run_name,
                                data_dir=data_dir,
                                project_root=project_root,
                                eval_reports_dir=eval_reports_dir,
                                mlflow_tracking_uri=mlflow_tracking_uri)
    orchestrator.run()



#! /home/vafaeisa/miniconda3/bin/python

import argparse
import os
from pathlib import Path
import subprocess

import pyslurm


# Make sure to clone/fetch your repository
SYNC_SCRIPT_PATH = '/home/vafaeisa/sync_mlflow.sh'

USER = os.getenv('USER')
MAX_JOBS = 10
mj = os.getenv('MAX_JOBS')
if mj is not None:
    MAX_JOBS = int(mj)


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('--project_root',
                        type=str,
                        help='**absolute** directory of the project repository.',
                        required=True)

    parser.add_argument('--run_name',
                        type=str,
                        help='repo_root_dir/runs/run_name',
                        required=True)

    parser.add_argument('--email',
                        type=str,
                        help='your job-events will be sent to this address',
                        required=True)

    parser.add_argument('--data_dir',
                        type=str,
                        help='**absolute** directory of the dataset',
                        required=False)

    parser.add_argument('--gpu_type',
                        type=str,
                        help='either p100, v100 or t4',
                        required=False,
                        default='p100')

    parser.add_argument('--hours',
                        type=int,
                        help='minimum required hours',
                        required=False,
                        default=2)

    parser.add_argument('--mem',
                        type=int,
                        help='minimum required memory in megabytes',
                        required=False,
                        default=12288)

    parser.add_argument('--n_cpu',
                        type=int,
                        help='cpus to use',
                        required=False,
                        default=1)

    return parser.parse_args()


def main():
    args = parse_args()

    jobs = pyslurm.job().find_user(USER)
    if len(jobs) >= MAX_JOBS:
        print(f'job queue is full, try again later.')
    else:
        project_root = Path(args.project_root).absolute()
        assert project_root.is_dir(), f"{project_root} is not a directory."

        run_name = args.run_name
        run_dir = project_root.joinpath('runs').joinpath(run_name)
        assert run_dir.is_dir(), f'run directory {run_dir} does not exist'

        job_name = f'{project_root.name}_{args.run_name}'

        email = args.email
        memory = args.mem
        hours = args.hours
        gpu_type = args.gpu_type
        n_cpu = args.n_cpu

        data_dir = args.data_dir
        if data_dir is not None:
            assert Path(data_dir).is_dir(), f'data directory {data_dir} does not exist'

        job_script_path = run_dir.joinpath(job_name + '.job')
        err_log_path = run_dir.joinpath(f'log-{job_name}.err')
        out_log_path = run_dir.joinpath(f'log-{job_name}.out')

        with open(job_script_path, 'w') as f:
            f.write('#!/bin/env bash\n\n')

            f.write('#SBATCH --get-user-env\n')
            f.write(f'#SBATCH --gres gpu:{gpu_type}:1\n')
            f.write(f'#SBATCH --job-name {job_name}\n')
            f.write(f'#SBATCH --time {hours}:00:00\n')
            f.write(f'#SBATCH --output {out_log_path}\n')
            f.write(f'#SBATCH --error {err_log_path}\n')
            f.write(f'#SBATCH --mem {memory}\n')
            f.write(f'#SBATCH --mail-user {email}\n')
            f.write(f'#SBATCH --cpus-per-task={n_cpu}\n')
            f.write(f'#SBATCH --mail-type ALL\n\n')

            # f.write(f'module load cuda/11.0\n')
            # f.write(f'module load cudnn\n\n')

            f.write('export CUDA_HOME=/cvmfs/soft.computecanada.ca/easybuild/software/2020/Core/cudacore/11.2.2\n')
            f.write('export PATH=$CUDA_HOME/bin${PATH:+:${PATH}}\n')
            # f.write('export LD_LIBRARY_PATH=$CUDA_HOME/lib64${LD_LIBRARY_PATH:+:${LD_LIBRARY_PATH\n\n')
            f.write('export CUDA_HDIR=/cvmfs/soft.computecanada.ca/easybuild/software/2020/Core/cudacore/11.2.2\n')
            f.write('export LD_LIBRARY_PATH=${CUDA_HOME}/nvvm/libdevice${LD_LIBRARY_PATH:+:${LD_LIBRARY_PATH}}\n\n')

            f.write(f'cd {project_root}\n')

            if data_dir is None:  # Use config.data_dir
                f.write(f'train --run_name {run_name}\n --project_root {project_root}\n')
            else:
                f.write(f'train --run_name {run_name} --data_dir {str(data_dir)} --project_root {project_root}\n')

        # os.system(f'sbatch {job_script_path}')
        submit_cmd = ['sbatch',  str(job_script_path)]
        res = subprocess.check_output(submit_cmd).strip()
        if not res.startswith(b"Submitted batch"):
            print(f'could not submit the batch: {res}')
        else:
            job_id = int(res.split()[-1])
            print(f'submitted job {job_id}')
            # os.system(f'strigger --set --jobid={job_id} --fini --program={SYNC_SCRIPT_PATH}')


if __name__ == '__main__':
    main()



import pytest
from pathlib import Path
from pydoc import locate
import numpy as np

from abstractions.utils import load_config_file
from abstractions import DataLoaderBase, AugmentorBase, PreprocessorBase, ModelBuilderBase, EvaluatorBase
from types import FunctionType


@pytest.fixture(scope='module')
def component_holder():
    return {}


@pytest.fixture()
def run_config(pytestconfig):
    config_path = pytestconfig.getoption('config_path')
    if config_path is None:
        run_name = pytestconfig.getoption('run_name')
        run_dir = Path('runs').joinpath(run_name)
        config = load_config_file(list(run_dir.glob('*.yaml'))[0])
    else:
        config = load_config_file(config_path)
    return config


# @pytest.mark.component
# @pytest.mark.dependency
# class TestConfigFile:


@pytest.mark.component
@pytest.mark.dependency
class TestLocateComponents:

    def test_locate_data_loader_class(self, run_config, component_holder):
        data_loader_class = locate(run_config.data_loader_class)
        assert data_loader_class is not None
        component_holder['data_loader_class'] = data_loader_class

    def test_locate_augmentor_class(self, run_config, component_holder):
        if not hasattr(run_config, 'augmentor_class'):
            assert True
        else:
            augmentor_class = locate(run_config.augmentor_class)
            assert augmentor_class is not None
            component_holder['augmentor_class'] = augmentor_class

    def test_locate_preprocessor_class(self, run_config, component_holder):
        preprocessor_class = locate(run_config.preprocessor_class)
        assert preprocessor_class is not None
        component_holder['preprocessor_class'] = preprocessor_class

    def test_locate_model_builder_class(self, run_config, component_holder):
        model_builder_class = locate(run_config.model_builder_class)
        assert model_builder_class is not None
        component_holder['model_builder_class'] = model_builder_class

    def test_locate_evaluator_class(self, run_config, component_holder):
        evaluator_class = locate(run_config.evaluator_class)
        assert evaluator_class is not None
        component_holder['evaluator_class'] = evaluator_class


@pytest.mark.component
class TestInitializeComponents:

    @pytest.mark.dependency(depends=['TestLocateComponents::test_locate_data_loader_class'])
    def test_initialize_data_loader(self, run_config, component_holder):
        data_dir = Path(run_config.data_dir)
        data_loader = component_holder.get('data_loader_class')(run_config, data_dir)
        assert isinstance(data_loader, DataLoaderBase)
        component_holder['data_loader'] = data_loader

    @pytest.mark.dependency(depends=['TestLocateComponents::test_locate_augmentor_class'])
    def test_initialize_augmentor(self, run_config, component_holder):
        augmentor_class = component_holder.get('augmentor_class')
        if augmentor_class is not None:
            augmentor = augmentor_class(run_config)
            assert isinstance(augmentor, AugmentorBase)
            component_holder['augmentor'] = augmentor

    @pytest.mark.dependency(depends=['TestLocateComponents::test_locate_preprocessor_class'])
    def test_initialize_preprocessor(self, run_config, component_holder):
        preprocessor = component_holder.get('preprocessor_class')(run_config)
        assert isinstance(preprocessor, PreprocessorBase)
        component_holder['preprocessor'] = preprocessor

    @pytest.mark.dependency(depends=['TestLocateComponents::test_locate_model_builder_class'])
    def test_initialize_model_builder(self, run_config, component_holder):
        model_builder = component_holder.get('model_builder_class')(run_config)
        assert isinstance(model_builder, ModelBuilderBase)
        component_holder['model_builder'] = model_builder

    @pytest.mark.dependency(depends=['TestLocateComponents::test_locate_evaluator_class'])
    def test_initialize_evaluator(self, run_config, component_holder):
        evaluator = component_holder.get('evaluator_class')(run_config)
        assert isinstance(evaluator, EvaluatorBase)
        component_holder['evaluator'] = evaluator


@pytest.mark.component
class TestDataLoader:

    # ---------- Train Data ----------
    @pytest.mark.dependency(depends=['TestInitializeComponents::test_initialize_data_loader'])
    def test_create_training_generator(self, run_config, component_holder):
        data_loader = component_holder['data_loader']
        ret = data_loader.create_training_generator()
        assert len(ret) == 2

        train_data_gen, train_n = ret
        component_holder['train_data_gen'] = train_data_gen
        component_holder['train_n'] = train_n

    @pytest.mark.dependency(depends=['TestDataLoader::test_create_training_generator'])
    def test_train_n(self, run_config, component_holder):
        assert isinstance(component_holder['train_n'], int)

    @pytest.mark.dependency(depends=['TestDataLoader::test_create_training_generator'])
    def test_train_generator(self, run_config, component_holder):
        assert hasattr(component_holder['train_data_gen'], '__iter__')

    @pytest.mark.dependency(depends=['TestDataLoader::test_train_generator'])
    def test_train_gen_out(self, component_holder):
        gen = component_holder['train_data_gen']
        ret = next(iter(gen))
        assert len(ret) == 3

    # ---------- Validation Data ----------
    @pytest.mark.dependency(depends=['TestInitializeComponents::test_initialize_data_loader'])
    def test_create_validation_generator(self, run_config, component_holder):
        data_loader = component_holder['data_loader']
        ret = data_loader.create_validation_generator()
        assert len(ret) == 2

        validation_data_gen, validation_n = ret
        component_holder['validation_data_gen'] = validation_data_gen
        component_holder['validation_n'] = validation_n

    @pytest.mark.dependency(depends=['TestDataLoader::test_create_validation_generator'])
    def test_validation_n(self, run_config, component_holder):
        assert hasattr(component_holder['validation_data_gen'], '__iter__')

    @pytest.mark.dependency(depends=['TestDataLoader::test_create_validation_generator'])
    def test_validation_generator(self, run_config, component_holder):
        assert isinstance(component_holder['validation_n'], int)

    @pytest.mark.dependency(depends=['TestDataLoader::test_create_validation_generator'])
    def test_validation_gen_out(self, component_holder):
        gen = component_holder['validation_data_gen']
        ret = next(iter(gen))
        assert len(ret) == 3

    # ---------- Test Data ----------
    @pytest.mark.dependency(depends=['TestInitializeComponents::test_initialize_data_loader'])
    def test_create_evaluation_generator(self, run_config, component_holder):
        data_loader = component_holder['data_loader']
        ret = data_loader.create_test_generator()
        assert len(ret) == 2

        evaluation_data_gen, evaluation_n = ret
        component_holder['evaluation_data_gen'] = evaluation_data_gen
        component_holder['evaluation_n'] = evaluation_n

    @pytest.mark.dependency(depends=['TestDataLoader::test_create_evaluation_generator'])
    def test_evaluation_n(self, run_config, component_holder):
        assert hasattr(component_holder['evaluation_data_gen'], '__iter__')

    @pytest.mark.dependency(depends=['TestDataLoader::test_create_evaluation_generator'])
    def test_evaluation_generator(self, run_config, component_holder):
        assert isinstance(component_holder['evaluation_n'], int)

    @pytest.mark.dependency(depends=['TestDataLoader::test_create_evaluation_generator'])
    def test_evaluation_gen_out(self, component_holder):
        gen = component_holder['evaluation_data_gen']
        ret = next(iter(gen))
        assert len(ret) == 3


@pytest.mark.component
class TestAugmentor:

    # ---------- Train Augmentation ----------
    @pytest.mark.dependency(depends=['TestDataLoader::test_train_generator'])
    def test_add_train_augmentation(self, run_config, component_holder):
        augmentor = component_holder.get('augmentor')
        if augmentor is None:
            assert True
        else:
            if run_config.do_train_augmentation:
                data_gen = augmentor.add_augmentation(component_holder['train_data_gen'])
                assert hasattr(data_gen, '__iter__')
                component_holder['train_data_gen'] = data_gen

    @pytest.mark.dependency(depends=['TestAugmentor::test_add_train_augmentation'])
    def test_train_augmentation_out(self, run_config, component_holder):
        train_data_gen = component_holder['train_data_gen']
        ret = next(iter(train_data_gen))
        assert len(ret) == 3

    # ---------- Validation Augmentation ----------
    @pytest.mark.dependency(depends=['TestDataLoader::test_validation_generator'])
    def test_add_validation_augmentation(self, run_config, component_holder):
        augmentor = component_holder.get('augmentor')
        if augmentor is None:
            assert True
        else:
            if run_config.do_validation_augmentation:
                data_gen = augmentor.add_validation_augmentation(component_holder['validation_data_gen'])
                assert hasattr(data_gen, '__iter__')
                component_holder['validation_data_gen'] = data_gen

    @pytest.mark.dependency(depends=['TestAugmentor::test_add_validation_augmentation'])
    def test_validation_augmentation_out(self, run_config, component_holder):
        data_gen = component_holder['validation_data_gen']
        ret = next(iter(data_gen))
        assert len(ret) == 3


@pytest.mark.component
class TestPreprocessor:

    # ---------- Train Preprocessing ----------
    @pytest.mark.dependency(depends=['TestAugmentor::test_add_train_augmentation'])
    def test_add_train_preprocessing(self, run_config, component_holder):
        preprocessor = component_holder['preprocessor']
        train_data_gen = component_holder['train_data_gen']
        train_n = component_holder['train_n']

        ret = preprocessor.add_preprocess(train_data_gen, train_n)
        assert len(ret) == 2

        train_gen, n_iter_train = ret
        component_holder['train_data_gen'] = train_gen
        component_holder['n_iter_train'] = n_iter_train

    @pytest.mark.dependency(depends=['TestPreprocessor::test_add_train_preprocessing'])
    def test_train_preprocess_generator(self, run_config, component_holder):
        assert hasattr(component_holder['train_data_gen'], '__iter__')

    @pytest.mark.dependency(depends=['TestPreprocessor::test_add_train_preprocessing'])
    def test_train_preprocess_n_iter(self, run_config, component_holder):
        assert isinstance(component_holder['n_iter_train'], int)

    @pytest.mark.dependency(depends=['TestPreprocessor::test_train_preprocess_generator'])
    def test_train_gen_out(self, run_config, component_holder):
        train_gen = component_holder['train_data_gen']
        ret = next(iter(train_gen))
        assert len(ret) == 3

        x_batch, y_batch, w_batch = ret
        component_holder['train_gen_xb_shape'] = tuple(x_batch.shape)
        component_holder['train_gen_yb_shape'] = tuple(y_batch.shape)
        component_holder['train_gen_wb'] = w_batch

    @pytest.mark.dependency(depends=['TestPreprocessor::test_train_gen_out'])
    def test_train_gen_x_shape(self, run_config, component_holder):
        batch_size = int(run_config.batch_size)
        input_h = int(run_config.input_height)
        input_w = int(run_config.input_width)
        x_b_shape = component_holder['train_gen_xb_shape']

        assert tuple(x_b_shape)[:3] == (batch_size, input_h, input_w)
        component_holder['input_channels'] = tuple(x_b_shape)[-1]

    @pytest.mark.dependency(depends=['TestPreprocessor::test_train_gen_out'])
    def test_train_gen_y_batch_size(self, run_config, component_holder):
        assert component_holder['train_gen_yb_shape'][0] == int(run_config.batch_size)

    @pytest.mark.dependency(depends=['TestPreprocessor::test_train_gen_out'])
    def test_train_gen_w_batch_size(self, run_config, component_holder):
        w_batch = component_holder['train_gen_wb']
        assert len(w_batch) == int(run_config.batch_size)

    @pytest.mark.dependency(depends=['TestPreprocessor::test_train_gen_out'])
    def test_train_gen_w_is_iterable(self, run_config, component_holder):
        w_batch = component_holder['train_gen_wb']
        assert hasattr(w_batch[0],
                       '__iter__'), 'elements of the weights_batch are not iterables, add a new axis to each element.'

    # ---------- Validation Preprocessing ----------
    @pytest.mark.dependency(depends=['TestAugmentor::test_add_validation_augmentation'])
    def test_add_validation_preprocessing(self, run_config, component_holder):
        preprocessor = component_holder['preprocessor']
        data_gen = component_holder['validation_data_gen']
        n = component_holder['validation_n']

        ret = preprocessor.add_preprocess(data_gen, n)
        assert len(ret) == 2

        gen, n_iter = ret
        component_holder['validation_data_gen'] = gen
        component_holder['n_iter_validation'] = n_iter

    @pytest.mark.dependency(depends=['TestPreprocessor::test_add_validation_preprocessing'])
    def test_validation_preprocess_generator(self, run_config, component_holder):
        assert hasattr(component_holder['validation_data_gen'], '__iter__')

    @pytest.mark.dependency(depends=['TestPreprocessor::test_add_validation_preprocessing'])
    def test_validation_preprocess_n_iter(self, run_config, component_holder):
        assert isinstance(component_holder['n_iter_validation'], int)

    @pytest.mark.dependency(depends=['TestPreprocessor::test_validation_preprocess_generator'])
    def test_validation_gen_out(self, run_config, component_holder):
        gen = component_holder['validation_data_gen']
        ret = next(iter(gen))
        assert len(ret) == 3

        x_batch, y_batch, w_batch = ret
        component_holder['validation_gen_xb_shape'] = tuple(x_batch.shape)
        component_holder['validation_gen_yb_shape'] = tuple(y_batch.shape)
        component_holder['validation_gen_wb'] = w_batch

    @pytest.mark.dependency(depends=['TestPreprocessor::test_validation_gen_out'])
    def test_validation_gen_x_shape(self, run_config, component_holder):
        batch_size = int(run_config.batch_size)
        input_h = int(run_config.input_height)
        input_w = int(run_config.input_width)
        x_b_shape = component_holder['validation_gen_xb_shape']

        assert tuple(x_b_shape)[:3] == (batch_size, input_h, input_w)
        component_holder['input_channels_validation'] = tuple(x_b_shape)[-1]

    @pytest.mark.dependency(depends=['TestPreprocessor::test_validation_gen_out'])
    def test_validation_gen_y_batch_size(self, run_config, component_holder):
        assert component_holder['validation_gen_yb_shape'][0] == int(run_config.batch_size)

    @pytest.mark.dependency(depends=['TestPreprocessor::test_validation_gen_out'])
    def test_validation_gen_w_batch_size(self, run_config, component_holder):
        w_batch = component_holder['validation_gen_wb']
        assert len(w_batch) == int(run_config.batch_size)

    @pytest.mark.dependency(depends=['TestPreprocessor::test_validation_gen_out'])
    def test_validation_gen_w_is_iterable(self, run_config, component_holder):
        w_batch = component_holder['validation_gen_wb']
        assert hasattr(w_batch[0],
                       '__iter__'), 'elements of the weights_batch are not iterables, add a new axis to each element.'

    # ---------- Evaluation Preprocessing ----------
    @pytest.mark.dependency(depends=['TestDataLoader::test_evaluation_generator'])
    def test_add_evaluation_preprocessing(self, run_config, component_holder):
        preprocessor = component_holder['preprocessor']
        data_gen = component_holder['evaluation_data_gen']
        n = component_holder['evaluation_n']

        # ret = preprocessor.add_preprocess(data_gen, n)
        gen = preprocessor.add_image_preprocess(data_gen)
        gen = preprocessor.add_label_preprocess(gen)
        ret = preprocessor.batchify(gen, n)

        assert len(ret) == 2

        gen, n_iter = ret
        component_holder['evaluation_data_gen'] = gen
        component_holder['n_iter_evaluation'] = n_iter

    @pytest.mark.dependency(depends=['TestPreprocessor::test_add_evaluation_preprocessing'])
    def test_evaluation_preprocess_generator(self, run_config, component_holder):
        assert hasattr(component_holder['evaluation_data_gen'], '__iter__')

    @pytest.mark.dependency(depends=['TestPreprocessor::test_add_evaluation_preprocessing'])
    def test_evaluation_preprocess_n_iter(self, run_config, component_holder):
        assert isinstance(component_holder['n_iter_evaluation'], int)

    @pytest.mark.dependency(depends=['TestPreprocessor::test_evaluation_preprocess_generator'])
    def test_evaluation_gen_out(self, run_config, component_holder):
        gen = component_holder['evaluation_data_gen']
        ret = next(iter(gen))
        assert len(ret) == 3

        x_batch, y_batch, w_batch = ret
        component_holder['evaluation_gen_xb_shape'] = tuple(x_batch.shape)
        component_holder['evaluation_gen_yb_shape'] = tuple(y_batch.shape)
        component_holder['evaluation_gen_id'] = w_batch

    @pytest.mark.dependency(depends=['TestPreprocessor::test_evaluation_gen_out'])
    def test_evaluation_gen_x_shape(self, run_config, component_holder):
        batch_size = int(run_config.batch_size)
        input_h = int(run_config.input_height)
        input_w = int(run_config.input_width)
        x_b_shape = component_holder['evaluation_gen_xb_shape']

        assert tuple(x_b_shape)[:3] == (batch_size, input_h, input_w)
        component_holder['input_channels_evaluation'] = tuple(x_b_shape)[-1]

    @pytest.mark.dependency(depends=['TestPreprocessor::test_evaluation_gen_out'])
    def test_evaluation_gen_y_batch_size(self, run_config, component_holder):
        assert component_holder['evaluation_gen_yb_shape'][0] == int(run_config.batch_size)

    @pytest.mark.dependency(depends=['TestPreprocessor::test_evaluation_gen_out'])
    def test_evaluation_gen_w_batch_size(self, run_config, component_holder):
        id_batch = component_holder['evaluation_gen_id']
        assert len(id_batch) == int(run_config.batch_size)

    # @pytest.mark.dependency(depends=['TestPreprocessor::test_evaluation_gen_out'])
    # def test_evaluation_gen_w_is_iterable(self, run_config, component_holder):
    #     w_batch = component_holder['evaluation_gen_wb']
    #     assert hasattr(w_batch[0],
    #                    '__iter__'), 'elements of the weights_batch are not iterables, add a new axis to each element.'


@pytest.mark.component
class TestModelBuilder:

    @pytest.mark.dependency(depends=['TestPreprocessor::test_train_preprocess_generator'])
    def test_get_compiled_model(self, run_config, component_holder):
        model_builder = component_holder['model_builder']
        compiled_model = model_builder.get_compiled_model()
        assert True  # TODO: check for being a tensorflow.keras model

        component_holder['compiled_model'] = compiled_model

    @pytest.mark.dependency(depends=['TestModelBuilder::test_get_compiled_model',
                                     'TestPreprocessor::test_train_preprocess_generator', ])
    def test_model_train_gen_compatibility(self, run_config, component_holder):
        compiled_model = component_holder['compiled_model']
        _, input_h, input_w, n_channels = compiled_model.input_shape

        train_gen = component_holder['train_data_gen']
        x_b, y_b, w_b = next(iter(train_gen))

        compiled_model.evaluate(x=x_b, y=y_b, sample_weight=w_b)
        assert True

        # assert input_h == int(run_config.input_height)
        # assert input_w == int(run_config.input_width)
        # assert n_channels == int(component_holder['input_channels'])

    @pytest.mark.dependency(depends=['TestModelBuilder::test_get_compiled_model',
                                     'TestPreprocessor::test_validation_preprocess_generator', ])
    def test_model_validation_gen_compatibility(self, run_config, component_holder):
        compiled_model = component_holder['compiled_model']
        _, input_h, input_w, n_channels = compiled_model.input_shape

        gen = component_holder['validation_data_gen']
        x_b, y_b, w_b = next(iter(gen))

        compiled_model.evaluate(x=x_b, y=y_b, sample_weight=w_b)
        assert True

        # assert input_h == int(run_config.input_height)
        # assert input_w == int(run_config.input_width)
        # assert n_channels == int(component_holder['input_channels'])


@pytest.mark.component
class TestEvaluation:

    @pytest.mark.dependency(depends=['TestInitializeComponents::test_initialize_evaluator',
                                     'TestPreprocessor::test_evaluation_gen_out'])
    def test_eval_funcs_on_evaluation_gen(self, component_holder):
        evaluator = component_holder['evaluator']
        eval_funcs = evaluator.get_eval_funcs()

        compiled_model = component_holder['compiled_model']
        data_gen = component_holder['evaluation_data_gen']
        x_b, y_b, _ = next(iter(data_gen))
        x_sample = x_b[0]
        y_sample = y_b[0]
        y_pred = compiled_model.predict(np.expand_dims(x_sample, axis=0))[0]

        failed_funcs = list()
        for f_name, f in eval_funcs.items():
            if not (isinstance(f_name, str) or isinstance(f, FunctionType)):
                failed_funcs.append(f_name)
            else:
                try:
                    f(y_sample, y_pred)
                except Exception as e:
                    failed_funcs.append(f_name)

        if any(failed_funcs):
            pytest.fail(f'failed {len(failed_funcs)}/{len(eval_funcs)}. failed funcs are {failed_funcs}')

    def test_eval_funcs_on_validation_gen(self, component_holder):
        evaluator = component_holder['evaluator']
        eval_funcs = evaluator.get_eval_funcs()

        compiled_model = component_holder['compiled_model']
        data_gen = component_holder['validation_data_gen']
        x_b, y_b, _ = next(iter(data_gen))
        x_sample = x_b[0]
        y_sample = y_b[0]
        y_pred = compiled_model.predict(np.expand_dims(x_sample, axis=0))[0]

        failed_funcs = list()
        for f_name, f in eval_funcs.items():
            if not (isinstance(f_name, str) or isinstance(f, FunctionType)):
                failed_funcs.append(f_name)
            else:
                try:
                    f(y_sample, y_pred)
                except Exception as e:
                    failed_funcs.append(f_name)

        if any(failed_funcs):
            pytest.fail(f'failed {len(failed_funcs)}/{len(eval_funcs)}. failed funcs are {failed_funcs}')

    # @pytest.mark.dependency(depends=['TestEvaluation::test_get_eval_funcs'])
    # def test_eval_funcs(self, component_holder):
    #     compiled_model = component_holder['compiled_model']
    #     eval_funcs = component_holder['eval_funcs']
    #     data_gen = component_holder['train_data_gen']
    #     x_b, y_b, _ = next(iter(data_gen))
    #     x_sample = x_b[0]
    #     y_sample = y_b[0]
    #     y_pred = compiled_model.predict(np.expand_dims(x_sample, axis=0))[0]
    #
    #     for f_name, f in eval_funcs.items():
    #         try:
    #             f(y_sample, y_pred)
    #         except Exception as e:
    #             pytest.fail(f'eval func {f_name} failed with exception {e}')


@pytest.mark.component
class TestTraining:

    @pytest.mark.dependency(depends=['TestEvaluation::test_eval_funcs_on_evaluation_gen',
                                     'TestEvaluation::test_eval_funcs_on_validation_gen'])
    def test_model_training(self, run_config, component_holder):
        """Simple model fit, with 3 batches per epoch, for 3 epochs."""

        model = component_holder['compiled_model']
        train_gen = component_holder['train_data_gen']
        val_gen = component_holder['validation_data_gen']

        x_tr, y_tr, w_tr = next(iter(train_gen))
        x_val, y_val, w_val = next(iter(val_gen))

        initial_tr_loss = model.evaluate(x_tr, y_tr, sample_weight=w_tr, return_dict=True)['loss']
        initial_val_loss = model.evaluate(x_val, y_val, sample_weight=w_val, return_dict=True)['loss']
        # train_iter = 3  # component_holder['n_iter_train']
        # val_iter = 3  # component_holder['n_iter_validation']

        epochs = 3

        history = model.fit(x_tr,
                            y_tr,
                            epochs=epochs,
                            sample_weight=w_tr,
                            validation_data=(x_val, y_val, w_val))
        assert True

        component_holder['training_history'] = history.history
        component_holder['initial_training_loss'] = initial_tr_loss
        component_holder['initial_validation_loss'] = initial_val_loss

    @pytest.mark.dependency(depends=['TestTraining::test_model_training'])
    def test_model_loss_is_decreasing(self, component_holder):
        model_loss = component_holder['training_history']['loss']
        assert model_loss[-1] < model_loss[0]

    @pytest.mark.dependency(depends=['TestTraining::test_model_training'])
    def test_model_val_loss_is_decreasing(self, component_holder):
        model_loss = component_holder['training_history']['val_loss']
        assert model_loss[-1] < model_loss[0]



import pytest
from pathlib import Path
from abstractions.utils import load_config_file, Struct


# @pytest.fixture()
# def run_config(pytestconfig):
#     run_name = pytestconfig.getoption("run_name")
#     run_dir = Path('runs').joinpath(run_name)
#     config = load_config_file(list(run_dir.glob('*.yaml'))[0])
#     return config

def get_configs():
    config_files = list()
    for config_path in Path('runs').rglob('*.yaml'):
        config_files.append(config_path)
    return config_files


@pytest.mark.init
def test_requirements():
    """Tests for main requirements."""

    packages = [
        "tensorflow>=2.7.0",
        "pandas",
        "scikit-learn",
        "scikit-image",
        "scipy",
        "matplotlib",
        "mlflow",
        "abstractions-aimedic",
        "SimpleITK",
        "pyyaml",
        "albumentations",
        "tqdm",
        "pytest",
        "pytest-dependency"
    ]

    # project_dir = Path(__file__).parent.parent
    # print(project_dir)

    # with open(project_dir.joinpath("requirements.txt"), "r") as f:
    with open('requirements.txt', 'r') as f:
        reqs = f.read().split()
        assert all(item in reqs for item in packages)


@pytest.mark.init
@pytest.mark.parametrize('config_path', get_configs())
def test_required_config_fields(config_path):
    """Testing ``required`` config parameters"""

    run_config = load_config_file(config_path)

    assert str(run_config.input_height).isalnum()
    assert str(run_config.input_width).isalnum()
    assert isinstance(run_config.src_code_path, str)
    assert isinstance(run_config.data_dir, str)
    assert isinstance(run_config.data_loader_class, str)
    assert isinstance(run_config.preprocessor_class, str)
    assert isinstance(run_config.model_builder_class, str)
    assert isinstance(run_config.evaluator_class, str)

    assert isinstance(run_config.export, Struct)
    assert hasattr(run_config.export, 'metric')
    assert hasattr(run_config.export, 'mode')


@pytest.mark.init
def test_train_script_existence():
    """Testing whether train script exists in ``scripts`` or not"""

    train_path = Path('scripts').joinpath('train.py')
    assert train_path.is_file()
